# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     David Herreros   (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import pyworkflow.utils as pwutils
from pyworkflow.object import Float
from pyworkflow.protocol import params

from pwem.objects import Volume, SetOfParticlesFlex
from pwem.protocols import ProtFlexBase

from xmipp3.protocols.protocol_preprocess import XmippResizeHelper
from xmipp3.protocols.protocol_preprocess.protocol_preprocess import XmippProcessParticles, XmippProcessVolumes

import flexutils.constants as const

import xmipp3


def _getSize(imgSet):
    """ get the size of an object"""
    if isinstance(imgSet, Volume):
        Xdim = imgSet.getDim()[0]
    else:
        Xdim = imgSet.getDimensions()[0]
    return Xdim

def _getSampling(imgSet):
    """ get the sampling rate of an object"""
    samplingRate = imgSet.getSamplingRate()
    return samplingRate


class XmippProtCropResizeZernikeParticles(XmippProcessParticles, ProtFlexBase):
    """ Crop or resize a set of particles with Zernike3D coefficients associated """
    _label = 'crop/resize zernike particles'
    _inputLabel = 'particles'

    def __init__(self, **kwargs):
        XmippProcessParticles.__init__(self, **kwargs)

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label=pwutils.Message.LABEL_INPUT)

        form.addParam('inputParticles', params.PointerParam,
                      pointerClass='SetOfParticlesFlex',
                      label=pwutils.Message.LABEL_INPUT_PART, important=True)
        # Hook that should be implemented in subclasses
        self._defineProcessParams(form)

        __threads, __mpi = self._getDefaultParallel()

        form.addParallelSection(threads=__threads, mpi=__mpi)

    def _defineProcessParams(self, form):
        XmippResizeHelper._defineProcessParams(self, form)
        form.addParallelSection(threads=0, mpi=8)

    def _insertProcessStep(self):
        XmippResizeHelper._insertProcessStep(self)

    # --------------------------- STEPS functions ---------------------------------------------------
    def filterStep(self, isFirstStep, args):
        XmippResizeHelper.filterStep(self, self._ioArgs(isFirstStep) + args)

    def resizeStep(self, isFirstStep, args):
        XmippResizeHelper.resizeStep(self, self._ioArgs(isFirstStep) + args)

    def windowStep(self, isFirstStep, args):
        XmippResizeHelper.windowStep(self, self._ioArgs(isFirstStep) + args)

    def _preprocessOutput(self, output):
        """ We need to update the sampling rate of the
        particles if the Resize option was used.
        """
        inputParticles = self.inputParticles.get()
        self.inputHasAlign = inputParticles.hasAlignment()

        output.L1 = inputParticles.getFlexInfo().L1
        output.L2 = inputParticles.getFlexInfo().L2
        if self.doResize:
            output.setSamplingRate(self.samplingRate)
            output.getFlexInfo().Rmax = Float(self.factor * self.inputParticles.get().getFlexInfo().Rmax.get())

    def _updateItem(self, item, row):
        """ Update also the sampling rate and
        the alignment if needed.
        """
        XmippProcessParticles._updateItem(self, item, row)
        if self.doResize:
            if item.hasCoordinate():
                item.scaleCoordinate(self.factor)
            item.setSamplingRate(self.samplingRate)
            if self.inputHasAlign:
                item.getTransform().scaleShifts(self.factor)

            # Scale Zernike3D information
            z_clnm = self.factor * item.getZFlex()
            item.setZFlex(z_clnm)

            if hasattr(item.getFlexInfo(), "deformation"):
                deformation = item.getFlexInfo().deformation.get()
                item.getFlexInfo().deformation = Float(self.factor * deformation)


    # --------------------------- INFO functions ----------------------------------------------------
    def _summary(self):
        summary = []

        if not hasattr(self, 'outputParticles'):
            summary.append("Output images not ready yet.")
        else:
            sampling = _getSampling(self.outputParticles)
            size = _getSize(self.outputParticles)
            if self.doResize:
                summary.append(u"Output particles have a different sampling "
                               u"rate (pixel size): *%0.3f* Å/px" % sampling)
                summary.append("Resizing method: *%s*" %
                               self.getEnumText('resizeOption'))
            if self.doWindow:
                if self.getEnumText('windowOperation') == "crop":
                    summary.append("The particles were cropped.")
                else:
                    summary.append("The particles were windowed.")
                summary.append("New size: *%s* px" % size)
        return summary

    def _methods(self):

        if not hasattr(self, 'outputParticles'):
            return []

        methods = ["We took input particles %s of size %d " % (
        self.getObjectTag('inputParticles'), len(self.inputParticles.get()))]
        if self.doWindow:
            if self.getEnumText('windowOperation') == "crop":
                methods += ["cropped them"]
            else:
                methods += ["windowed them"]
        if self.doResize:
            outputParticles = getattr(self, 'outputParticles', None)
            if outputParticles is None or outputParticles.getDim() is None:
                methods += ["Output particles not ready yet."]
            else:
                methods += ['resized them to %d px using the "%s" method%s' %
                            (outputParticles.getDim()[0],
                             self.getEnumText('resizeOption'),
                             " in Fourier space" if self.doFourier else "")]
        if not self.doResize and not self.doWindow:
            methods += ["did nothing to them"]
        str = "%s and %s. Output particles: %s" % (", ".join(methods[:-1]),
                                                   methods[-1],
                                                   self.getObjectTag('outputParticles'))
        return [str]

    def _validate(self):
        errors = XmippResizeHelper._validate(self)

        inputParticles = self.inputParticles.get()

        inputParticles = self.inputParticles.get()
        if isinstance(inputParticles, SetOfParticlesFlex):
            if inputParticles.getFlexInfo().getProgName() != const.ZERNIKE3D:
                errors.append("The flexibility information associated with the particles is not "
                              "coming from the Zernike3D algorithm. Please, provide a set of particles "
                              "with the correct flexibility information.")

        return errors

    # --------------------------- UTILS functions ---------------------------------------------------
    def _ioArgs(self, isFirstStep):
        if isFirstStep:
            return "-i %s -o %s --save_metadata_stack %s --keep_input_columns " % (
            self.inputFn, self.outputStk, self.outputMd)
        else:
            return "-i %s " % self.outputStk

    def _filterArgs(self):
        return XmippResizeHelper._filterCommonArgs(self)

    def _resizeArgs(self):
        return XmippResizeHelper._resizeCommonArgs(self)

    def _windowArgs(self):
        return XmippResizeHelper._windowCommonArgs(self)

    def _getSetSize(self):
        """ get the size of SetOfParticles object"""
        imgSet = self.inputParticles.get()
        return _getSize(imgSet)

    def _getSetSampling(self):
        """ get the sampling rate of SetOfParticles object"""
        imgSet = self.inputParticles.get()
        return _getSampling(imgSet)

    def _getDefaultParallel(self):
        """ Return the default value for thread and MPI
        for the parallel section definition.
        """
        return (0, 1)


class XmippProtCropResizeZernikeVolumes(XmippProcessVolumes):
    """ Crop or resize a set of volumes with Zernike3D coefficients associated """
    _label = 'crop/resize zernike volumes'
    _inputLabel = 'volumes'

    def __init__(self, **kwargs):
        XmippProcessVolumes.__init__(self, **kwargs)

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label=pwutils.Message.LABEL_INPUT)

        form.addParam('inputVolumes', params.PointerParam, important=True,
                      label=pwutils.Message.LABEL_INPUT_VOLS, pointerClass='VolumeFlex, SetOfVolumesFlex',
                      help='Can be a density volume or a SetOfVolumesFlex')
        # Hook that should be implemented in subclasses
        self._defineProcessParams(form)
        form.addParallelSection(threads=2, mpi=1)

    def _defineProcessParams(self, form):
        XmippResizeHelper._defineProcessParams(self, form)

    def _insertProcessStep(self):
        XmippResizeHelper._insertProcessStep(self)

    # --------------------------- STEPS functions ---------------------------------------------------
    def filterStep(self, isFirstStep, args):
        self.runJob("xmipp_transform_filter", self._ioArgs(isFirstStep) + args, env=xmipp3.Plugin.getEnviron())

    def resizeStep(self, isFirstStep, args):
        self.runJob("xmipp_image_resize", self._ioArgs(isFirstStep) + args, env=xmipp3.Plugin.getEnviron())

    def windowStep(self, isFirstStep, args):
        self.runJob("xmipp_transform_window", self._ioArgs(isFirstStep) + args, env=xmipp3.Plugin.getEnviron())

    def _preprocessOutput(self, volumes):
        # We use the preprocess only whne input is a set
        # we do not use postprocess to setup correctly
        # the samplingRate before each volume is added
        if not self._isSingleInput():
            if self.doResize:
                volumes.setSamplingRate(self.samplingRate)

    def _postprocessOutput(self, volumes):
        # We use the postprocess only when input is a volume
        if self._isSingleInput():
            if self.doResize:
                volumes.setSamplingRate(self.samplingRate)
                # we have a new sampling so origin need to be adjusted
                iSampling = self.inputVolumes.get().getSamplingRate()
                oSampling = self.samplingRate
                xdim_i, ydim_i, zdim_i = self.inputVolumes.get().getDim()
                xdim_o, ydim_o, zdim_o = volumes.getDim()

                xOrig, yOrig, zOrig = \
                    self.inputVolumes.get().getShiftsFromOrigin()
                xOrig += (xdim_i * iSampling - xdim_o * oSampling) / 2.
                yOrig += (ydim_i * iSampling - ydim_o * oSampling) / 2.
                zOrig += (zdim_i * iSampling - zdim_o * oSampling) / 2.
                volumes.setShiftsInOrigin(xOrig, yOrig, zOrig)
                volumes.setSamplingRate(oSampling)

                # Scale Zernike3D information
                # FIXME: Esto HAY que arreglarlo (no guarda los valores)
                for item in volumes.iterItems(iterate=False):
                    item.getFlexInfo().Rmax = Float(self.factor * item.getFlexInfo().Rmax.get())
                    z_clnm = self.factor * item.getZFlex()
                    item.setZFlex(z_clnm)

    # --------------------------- INFO functions ----------------------------------------------------
    def _summary(self):
        summary = []

        if not hasattr(self, 'outputVol'):
            summary.append("Output volume(s) not ready yet.")
        else:
            sampling = _getSampling(self.outputVol)
            size = _getSize(self.outputVol)
            if self.doResize:
                summary.append(u"Output volume(s) have a different sampling "
                               u"rate (pixel size): *%0.3f* Å/px" % sampling)
                summary.append("Resizing method: *%s*" %
                               self.getEnumText('resizeOption'))
            if self.doWindow.get():
                if self.getEnumText('windowOperation') == "crop":
                    summary.append("The volume(s) were cropped.")
                else:
                    summary.append("The volume(s) were windowed.")
                summary.append("New size: *%s* px" % size)
        return summary

    def _methods(self):
        if not hasattr(self, 'outputVol'):
            return []

        if self._isSingleInput():
            methods = ["We took one volume"]
            pronoun = "it"
        else:
            methods = ["We took %d volumes" % self.inputVolumes.get().getSize()]
            pronoun = "them"
        if self.doWindow.get():
            if self.getEnumText('windowOperation') == "crop":
                methods += ["cropped %s" % pronoun]
            else:
                methods += ["windowed %s" % pronoun]
        if self.doResize:
            outputVol = getattr(self, 'outputVol', None)
            if outputVol is None or self.outputVol.getDim() is None:
                methods += ["Output volume not ready yet."]
            else:
                methods += ['resized %s to %d px using the "%s" method%s' %
                            (pronoun, self.outputVol.getDim()[0],
                             self.getEnumText('resizeOption'),
                             " in Fourier space" if self.doFourier else "")]
        if not self.doResize and not self.doWindow:
            methods += ["did nothing to %s" % pronoun]
            # TODO: does this case even work in the protocol?
        return ["%s and %s." % (", ".join(methods[:-1]), methods[-1])]

    def _validate(self):
        return XmippResizeHelper._validate(self)

    # --------------------------- UTILS functions ---------------------------------------------------
    def _ioArgs(self, isFirstStep):
        if isFirstStep:
            if self._isSingleInput():
                return "-i %s -o %s " % (self.inputFn, self.outputStk)
            else:
                return "-i %s -o %s --save_metadata_stack %s --keep_input_columns " % (
                self.inputFn, self.outputStk, self.outputMd)
        else:
            return "-i %s" % self.outputStk

    def _filterArgs(self):
        return XmippResizeHelper._filterCommonArgs(self)

    def _resizeArgs(self):
        return XmippResizeHelper._resizeCommonArgs(self)

    def _windowArgs(self):
        return XmippResizeHelper._windowCommonArgs(self)

    def _getSetSize(self):
        """ get the size of either Volume or SetOfVolumes objects"""
        imgSet = self.inputVolumes.get()
        size = _getSize(imgSet)
        return size

    def _getSetSampling(self):
        """ get the sampling rate of either Volume or SetOfVolumes objects"""
        imgSet = self.inputVolumes.get()
        samplingRate = _getSampling(imgSet)
        return samplingRate


# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import re
import glob
import numpy as np
from xmipp_metadata.image_handler import ImageHandler

from pyworkflow import NEW
from pyworkflow.protocol.params import PointerParam, FloatParam
import pyworkflow.utils as pwutils

from pwem.protocols import ProtAnalysis3D

import xmipp3

class ProtFlexAutoReference(ProtAnalysis3D):
    """ Automatic selection of best reference volume for Zernike3D analysis """

    _label = 'auto reference'
    _devStatus = NEW
    OUTPUT_PREFIX = 'selectedReference'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('maps', PointerParam, label="Maps to select reference from",
                      pointerClass='SetOfVolumes', important=True)
        form.addParam('targetResolution', FloatParam, label="Target resolution (A)", default=8.0,
                      help="In Angstroms, the images and the volume are rescaled so that this resolution is at "
                           "2/3 of the Fourier spectrum.")
        form.addParam('angSampling', FloatParam, default=5.0, label="Angular sampling",
                      help="Angular step determining how many combinations of (rot,tilt) will be "
                           "evaluated during the comparison")
        form.addParallelSection(threads=4, mpi=0)

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.computeDecisionMatrix)
        self._insertFunctionStep(self.createOutputStep)

    def createOutputStep(self):
        if not hasattr(self, 'decision_matrix'):
            files = sorted(glob.glob(self._getTmpPath('map_*.mrc')), key=self.numericalSort)
            num_files = len(files)
            ih = ImageHandler()
            self.decision_matrix = np.zeros([num_files, num_files])
            for idx in range(num_files - 1):
                imgFile = self._getExtraPath("Matrix_entry_image_%d_%d.xmp" % (idx, idx + 1))
                density_cmp_image = ih.read(imgFile).getData()
                sum_density_cmp = np.sum(density_cmp_image)
                if sum_density_cmp > 0.0:
                    self.decision_matrix[idx, idx + 1] = 1.0
                    self.decision_matrix[idx + 1, idx] = -1.0
                elif sum_density_cmp < 0.0:
                    self.decision_matrix[idx, idx + 1] = -1.0
                    self.decision_matrix[idx + 1, idx] = 1.0

        row_sum = np.abs(self.decision_matrix.sum(axis=0))
        volId_decision = np.argmax(row_sum) + 1
        maps = self.maps.get()
        outVol = maps[int(volId_decision)].clone()

        # Save new output
        name = self.OUTPUT_PREFIX
        args = {}
        args[name] = outVol
        self._defineOutputs(**args)
        self._defineSourceRelation(maps, outVol)

    # --------------------------- STEPS functions -----------------------------
    def computeDecisionMatrix(self):
        maps = self.maps.get()
        angSampling = self.angSampling.get()

        # Compute new dimensions
        newTs = self.getNewTs()
        newXDim = self.getNewDim(newTs)

        # Move maps to tmp path and compute corr image
        ih = ImageHandler()
        Xdim = maps[1].getDim()[0]
        self.newAngSampling = 360. / np.round(360. / angSampling)
        for idx, compareMap in enumerate(maps.iterItems()):
            inputFile = compareMap.getFileName()
            mapFile = self._getTmpPath('map_%d.mrc' % (idx + 1))
            if pwutils.getExt(inputFile) == ".mrc":
                inputFile += ":mrc"
            if Xdim != newXDim:
                self.runJob("xmipp_image_resize",
                            "-i %s -o %s --dim %d " % (inputFile,
                                                       mapFile,
                                                       newXDim),
                            numberOfMpi=1, env=xmipp3.Plugin.getEnviron())
            else:
                ih.convert(inputFile, mapFile)

        # Compute density comparison image
        files = sorted(glob.glob(self._getTmpPath('map_*.mrc')), key=self.numericalSort)
        num_files = len(files)
        self.decision_matrix = np.zeros([num_files, num_files])
        for idx in range(num_files - 1):
            file_1 = files[idx]
            file_2 = files[idx + 1]
            outFile = self._getExtraPath("Matrix_entry_image_%d_%d.xmp" % (idx, idx + 1))
            self.runJob("xmipp_compare_density",
                        "-v1 %s -v2 %s -o %s --degstep %f --thr %d "
                        % (file_1, file_2, outFile, self.newAngSampling, self.numberOfThreads.get()),
                        env=xmipp3.Plugin.getEnviron())

            # Fill matrix entry
            density_cmp_image = ih.read(outFile).getData()
            sum_density_cmp = np.sum(density_cmp_image)
            if sum_density_cmp > 0.0:
                self.decision_matrix[idx, idx + 1] = 1.0
                self.decision_matrix[idx + 1, idx] = -1.0
            elif sum_density_cmp < 0.0:
                self.decision_matrix[idx, idx + 1] = -1.0
                self.decision_matrix[idx + 1, idx] = 1.0

    # --------------------------- UTILS functions ----------------------------
    def getNewDim(self, Ts):
        map = self.maps.get()[1]
        Xdim = map.getXDim()
        return int(Xdim * map.getSamplingRate() / Ts)

    def getNewTs(self):
        map = self.maps.get()[1]
        targetResolution = self.targetResolution.get()
        Ts = map.getSamplingRate()
        newTs = targetResolution / 3.0
        newTs = max(Ts, newTs)
        return newTs

    def numericalSort(self, value):
        numbers = re.compile(r'(\d+)')
        parts = numbers.split(value)
        parts[1::2] = map(int, parts[1::2])
        return parts

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            for _, outRegions in self.iterOutputAttributes():
                summary.append("Found best reference")
        else:
            summary.append("Finding best reference in set...")
        return summary

    def _methods(self):
        return [
            "Automatic reference selection for Zernike3D analysis",
        ]



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


from pwem.viewers import showj
from pwem.viewers.showj import runJavaIJapp
from pyworkflow.gui import *


def launchIJForSelection(path, image):
    macroPath = os.path.join(path, "AutoSave_ROI.ijm")
    imageBaseName = os.path.basename(image)

    macro = r"""
    path = "%s";
    file = "%s"
    
    // --------- Initialize Roi Manager ---------
    roiManager("Draw");
    setTool("freehand");
    run("Fire");
    run("Scale to Fit");
    
    newClass = "Yes";
    outPath = path + file + ".txt";
    
    // --------- Load SetOfMeshes ---------
    if (File.exists(outPath)){
    group = 0;
    groups = loadMeshFile(outPath);
    numMeshes = roiManager("count");
    emptyOutFile(outPath);
    aux = editMeshes(groups, numMeshes, outPath);
    group = group + aux + 1;
    }
    else{
    emptyOutFile(outPath);
    group = 1;
    }
    
    // --------- Draw new Meshes and save them ---------
    roiManager("Reset");
    waitForRoi();
    saveMeshes(group, outPath);
    
    // --------- Close ImageJ ---------
    run("Quit");
    
    
    // --------- Functions Definition ---------        
    function waitForRoi(){
    waitForUser("Draw the desired ROIs\n\nThen click Ok");
    wait(50);
    while(roiManager("count")==0){
    waitForUser("Draw the desired ROIs\n\nThen click Ok");
    wait(50);
    }
    }
    
    function emptyOutFile(outPath){
    fid = File.open(outPath);
    File.close(fid);
    }
    
    function saveMeshes(class, outPath){
    string = "";
    meshes = roiManager("count");
    for (i=0; i<meshes; i++){
    roiManager("select", i);
    Stack.getPosition(channel, slice, frame);
    getSelectionCoordinates(xpoints, ypoints);
    for (j = 0; j < xpoints.length; j++) {
    string = string + "" + xpoints[j] + "," + ypoints[j] + "," + slice + "," + class + "\n";
    }
    class++;
    }
    File.append(string, outPath);
    }
    
    function loadMeshFile(meshPath){
    c = "";
    c = c + toHex(255);
    c = c + toHex(255);
    c = c + toHex(255);
    contents = split(File.openAsString(meshPath), "\n");
    xpoints = newArray();
    ypoints = newArray();
    groups = newArray();
    for (idx=0; idx < contents.length; idx++){
    values = split(contents[idx], ",");
    valuesNext = split(contents[idx+1], ",");
    if (contents[idx+1] == "") {
    valuesNext =  newArray(-1,-1,-1,-1);
    idx++;
    }
    xpoints = Array.concat(xpoints, values[0]);
    ypoints = Array.concat(ypoints, values[1]);
    if (values[3] != valuesNext[3]){
    xpoints = Array.concat(xpoints, xpoints[0]);
    ypoints = Array.concat(ypoints, ypoints[0]);
    groups = Array.concat(groups, values[3]);
    makeSelection("polyline", xpoints, ypoints);
    Roi.setName("Class " + values[3]);
    roiManager("add");
    count = roiManager("count");
    roiManager("select", count-1);
    roiManager("Set Color", c);
    xpoints = newArray();
    ypoints = newArray();
    }
    }
    return groups;
    }
    
    function editMeshes(classVect, numMeshes, outPath){
    waitForUser("Edit the input ROIs if needed\n\nThen click Ok");
    string = "";
    for (i=0; i<numMeshes; i++){
    roiManager("select", i);
    Stack.getPosition(channel, slice, frame);
    getSelectionCoordinates(xpoints, ypoints);
    for (j = 0; j < xpoints.length; j++) {
    string = string + "" + xpoints[j] + "," + ypoints[j] + "," + slice + "," + classVect[i] + "\n";
    groupEnd = classVect[i];
    }
    }
    File.append(string, outPath);
    return groupEnd;
    }
""" % (os.path.join(path, ''), os.path.splitext(imageBaseName)[0])
    macroFid = open(macroPath, 'w')
    macroFid.write(macro)
    macroFid.close()

    args = "-i %s -macro %s" % (image, macroPath)
    viewParams = {showj.ZOOM: 300}
    for key, value in viewParams.items():
        args = "%s --%s %s" % (args, key, value)

    app = "xmipp.ij.commons.XmippImageJ"

    runJavaIJapp(4, app, args).wait()


def lanchIJForViewing(self, path, image):
    self.macroPath = os.path.join(self.path, "View_ROI.ijm")
    imageBaseName = os.path.basename(image)

    roiFile = os.path.join(path, os.path.splitext(imageBaseName)[0] + ".txt")

    macro = r"""path = "%s";
file = "%s"
meshFile = "%s"

outPath = path + meshFile;

// --------- Load SetOfMeshes ---------
if (File.exists(outPath)){
loadMeshFile(outPath);
}

// --------- Functions Definition ---------    
function loadMeshFile(meshPath){
c = "";
c = c + toHex(255*random);
c = c + toHex(255*random);
c = c + toHex(255*random);
contents = split(File.openAsString(meshPath), "\n");
xpoints = newArray();
ypoints = newArray();
groups = newArray();
for (idx=0; idx < contents.length; idx++){
values = split(contents[idx], ",");
if (idx+1 < contents.length){
valuesNext = split(contents[idx+1], ",");
}
else{
valuesNext =  newArray(-1,-1,-1,-1);
}
xpoints = Array.concat(xpoints, values[0]);
ypoints = Array.concat(ypoints, values[1]);
if (values[2] != valuesNext[2]){
xpoints = Array.concat(xpoints, xpoints[0]);
ypoints = Array.concat(ypoints, ypoints[0]);
groups = Array.concat(groups, values[3]);
Stack.setSlice(values[2]);
makeSelection("polyline", xpoints, ypoints);
Roi.setName("Class " + values[3]);
Roi.setStrokeWidth(5);
roiManager("add");
count = roiManager("count");
roiManager("select", count-1);
roiManager("Set Color", c);
if (values[3] != valuesNext[3]){
c = "";
c = c + toHex(255*random);
c = c + toHex(255*random);
c = c + toHex(255*random);
}
xpoints = newArray();
ypoints = newArray();
}
}
return groups;
}
""" % (os.path.join(path, ''), os.path.splitext(imageBaseName)[0], roiFile)
    macroFid = open(self.macroPath, 'w')
    macroFid.write(macro)
    macroFid.close()

    args = "-i %s -macro %s" % (image, self.macroPath)
    viewParams = {showj.ZOOM: 50}
    for key, value in viewParams.items():
        args = "%s --%s %s" % (args, key, value)

    app = "xmipp.ij.commons.XmippImageJ"

    runJavaIJapp(4, app, args).wait()



# **************************************************************************
# *
# * Authors:     Amaya Jimenez Moreno (ajimenez@cnb.csic.es)
# *              David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
from xmipp_metadata.image_handler import ImageHandler

from pwem.protocols import ProtAnalysis3D
import pyworkflow.protocol.params as params
from pwem.objects import Volume
from pyworkflow import VERSION_2_0
from pyworkflow.object import Integer, String

import xmipp3

from pwem.objects import VolumeFlex

from flexutils.utils import readZernikeFile
import flexutils.constants as const
import flexutils


class ProtFlexVolumeDeformZernike3D(ProtAnalysis3D):
    """ Protocol for volume deformation based on Zernike3D. """
    _label = 'volume deform - Zernike3D'
    _lastUpdateVersion = VERSION_2_0

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addHidden(params.USE_GPU, params.BooleanParam, default=True,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                           Select the one you want to use.")
        form.addHidden(params.GPU_LIST, params.StringParam, default='0',
                       expertLevel=params.LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        form.addParam('inputVolume', params.PointerParam, label="Input volume",
                      pointerClass='Volume',
                      help="Volume to be deformed")
        form.addParam('refVolume', params.PointerParam, label="Reference volume",
                      pointerClass='Volume',
                      help="Target volume input volume will be deformed to")
        form.addParam('sigma', params.NumericListParam, label="Multiresolution", default="1 2",
                      help="Perform the analysys comparing different filtered versions of the volumes")
        form.addParam('boxSize', params.IntParam, label="Downsample maps to this boxsize",
                      default=128,
                      help="Internally downsample the input and reference maps to speed up computations. "
                           "Downsampling is only applied internally, resulting volumes will be resampled to "
                           "the original box size.")
        form.addParam('l1', params.IntParam, default=3,
                      label='Zernike Degree',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Degree Zernike Polynomials of the deformation=1,2,3,...')
        form.addParam('l2', params.IntParam, default=2,
                      label='Harmonical Degree',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Degree Spherical Harmonics of the deformation=1,2,3,...')
        form.addParam('penalization', params.FloatParam, default=0.00025, label='Regularization',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Penalization to deformations (higher values penalize more the deformation).')
        form.addParallelSection(threads=4, mpi=0)


    def _createFilenameTemplates(self):
        """ Centralize how files are called """
        myDict = {
            'fnRefVol': self._getExtraPath('ref_volume.mrc'),
            'fnInputVol': self._getExtraPath('input_volume.mrc'),
            'fnInputFilt': self._getExtraPath('input_volume_filt.mrc'),
            'fnRefFilt': self._getExtraPath('ref_volume_filt.mrc'),
            'fnOutVol': self._getExtraPath('vol1DeformedTo2.mrc')
                 }
        self._updateFilenamesDict(myDict)

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._createFilenameTemplates()
        self._insertFunctionStep(self.convertInputStep)
        self._insertFunctionStep(self.deformStep)
        self._insertFunctionStep(self.convertOutputStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions ------------------------------
    def convertInputStep(self):
        fnInputVol = self._getFileName('fnInputVol')
        fnRefVol = self._getFileName('fnRefVol')

        XdimI = self.inputVolume.get().getDim()[0]
        XdimR = self.refVolume.get().getDim()[0]
        newXdim = self.boxSize.get()

        ih = ImageHandler()
        ih.convert(self.inputVolume.get().getFileName(), fnInputVol)
        if XdimI != newXdim:
            ImageHandler().scaleSplines(inputFn=fnInputVol, outputFn=fnInputVol, finalDimension=newXdim,
                                        overwrite=True)


        ih.convert(self.refVolume.get().getFileName(), fnRefVol)
        if XdimR != newXdim:
            ImageHandler().scaleSplines(inputFn=fnRefVol, outputFn=fnRefVol, finalDimension=newXdim,
                                        overwrite=True)


    def deformStep(self):
        fnRefVol = self._getFileName('fnRefVol')
        fnOutVol = self._getFileName('fnOutVol')
        fnRefMask = self._getExtraPath("ref_mask.mrc")
        fnOutMask = self._getExtraPath("ref_mask.mrc")
        fnZclnm = self._getExtraPath('Volumes_clnm.txt')

        self.alignMaps()

        mask_input = ImageHandler(fnOutVol).generateMask(iterations=50, boxsize=64, smoothStairEdges=False)
        ImageHandler().write(mask_input, fnOutMask, overwrite=True)
        mask_reference = ImageHandler(fnOutVol).generateMask(iterations=50, boxsize=64, smoothStairEdges=False)
        ImageHandler().write(mask_reference, fnRefMask, overwrite=True)

        params = ' -i %s -r %s -o %s --analyzeStrain --l1 %d --l2 %d --sigma "%s" --oroot %s --regularization %f' % \
                 (fnOutVol, fnRefVol, fnOutVol, self.l1.get(), self.l2.get(), self.sigma.get(),
                  self._getExtraPath('Volumes'), self.penalization.get())

        if self.useGpu.get():
            self.runJob("xmipp_cuda_volume_deform_sph", params, env=xmipp3.Plugin.getEnviron())
        else:
            if self.numberOfThreads.get() != 0:
                params = params + ' --thr %d' % self.numberOfThreads.get()
            self.runJob("xmipp_volume_deform_sph", params, env=xmipp3.Plugin.getEnviron())

        # Invert deformation field
        args = "--i %s --r %s --z_clnm %s --o %s" % \
               (fnRefMask, fnOutMask, fnZclnm, fnZclnm)
        program = "invert_zernike_field.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args)

    def convertOutputStep(self):
        inputVolume = self.inputVolume.get()
        xDim = inputVolume.getXDim()
        sr_i = inputVolume.getSamplingRate()
        newXDim = self.boxSize.get()
        fnOutVol = self._getFileName('fnOutVol')
        if newXDim != xDim:
            ImageHandler().scaleSplines(inputFn=fnOutVol, outputFn=fnOutVol,
                                        finalDimension=xDim,
                                        overwrite=True)
        ImageHandler().setSamplingRate(fnOutVol, sr_i)

    def createOutputStep(self):
        inputVolume = self.inputVolume.get()
        xDim = inputVolume.getXDim()
        newXDim = self.boxSize.get()
        sr_i = inputVolume.getSamplingRate()
        basis_params, z_clnm = readZernikeFile(self._getExtraPath('Volumes_clnm.txt'))
        z_clnm *= xDim / newXDim

        # Create input map mask
        zernike_mask = self._getExtraPath("zernike_mask.mrc")
        mask_input = ImageHandler(inputVolume.getFileName()).generateMask(iterations=50, boxsize=64, smoothStairEdges=False)
        ImageHandler().write(mask_input, zernike_mask, overwrite=True)

        # Deformed volume
        out_vol = Volume()
        out_vol.setLocation(self._getFileName('fnOutVol'))
        out_vol.setSamplingRate(sr_i)

        # Flex volume
        vol_flex = VolumeFlex(progName=const.ZERNIKE3D)
        vol_flex.setSamplingRate(sr_i)
        vol_flex.setFileName(inputVolume.getFileName())
        vol_flex.getFlexInfo().L1 = Integer(self.l1.get())
        vol_flex.getFlexInfo().L2 = Integer(self.l2.get())
        vol_flex.getFlexInfo().Rmax = Integer(0.5 * xDim)
        vol_flex.getFlexInfo().refMap = String(inputVolume.getFileName())
        vol_flex.getFlexInfo().refMask = String(zernike_mask)
        vol_flex.setZFlex(z_clnm[0])

        self._defineOutputs(deformed=out_vol, flexVolume=vol_flex)
        self._defineSourceRelation(self.inputVolume, out_vol)
        self._defineSourceRelation(self.inputVolume, vol_flex)

    def alignMaps(self):
        fnInputVol = self._getFileName('fnInputVol')
        fnInputFilt = self._getFileName('fnInputFilt')
        fnRefVol = self._getFileName('fnRefVol')
        fnRefFilt = self._getFileName('fnRefFilt')
        fnOutVol = self._getFileName('fnOutVol')

        # Filter the volumes to improve alignment quality
        params = " -i %s -o %s --fourier real_gaussian 2" % (fnInputVol, fnInputFilt)
        self.runJob("xmipp_transform_filter", params, env=xmipp3.Plugin.getEnviron())
        params = " -i %s -o %s --fourier real_gaussian 2" % (fnRefVol, fnRefFilt)
        self.runJob("xmipp_transform_filter", params, env=xmipp3.Plugin.getEnviron())

        # Find transformation needed to align the volumes
        params = ' --i1 %s --i2 %s --local --dontScale ' \
                 '--copyGeo %s' % \
                 (fnRefFilt, fnInputFilt, self._getExtraPath("geo.txt"))
        self.runJob("xmipp_volume_align", params, env=xmipp3.Plugin.getEnviron())

        # Apply transformation of filtered volume to original volume
        with open(self._getExtraPath("geo.txt"), 'r') as file:
            geo_str = file.read().replace('\n', ',')
        params = " -i %s -o %s --matrix %s" % (fnInputVol, fnOutVol, geo_str)
        self.runJob("xmipp_transform_geometry", params, env=xmipp3.Plugin.getEnviron())

    # ------------------------- VALIDATE functions -----------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        l1 = self.l1.get()
        l2 = self.l2.get()
        if (l1 - l2) < 0:
            errors.append('Zernike degree must be higher than '
                          'SPH degree.')
        return errors


# **************************************************************************
# *
# * Authors:  David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np

from pyworkflow.viewer import DESKTOP_TKINTER, WEB_DJANGO, ProtocolViewer
import pyworkflow.protocol.params as params

from flexutils.protocols.xmipp.protocol_interactive_flex_consensus import TensorflowProtInteractiveFlexConsensus
from flexutils.viewers.consensus_viewers.viewer_interactive_hist import InteractiveHist


class XmippFlexConsensusView(ProtocolViewer):
    """ Interactive FlexConsensus filtering """
    _label = 'viewer FlexConsensus'
    _targets = [TensorflowProtInteractiveFlexConsensus]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]
    _choices = ["Mean matched", "Entropy matched"]

    def __init__(self, **kwargs):
        ProtocolViewer.__init__(self, **kwargs)

    def _defineParams(self, form):
        form.addSection(label='Show deformation')
        form.addParam('histChoice', params.EnumParam,
                      choices=self._choices, default=0,
                      label='Error histogram to display', display=params.EnumParam.DISPLAY_COMBO,
                      help="\t Mean matched: Error distribution matched to the lowest error distribution " \
                           "found by FlexConsensus\n" \
                           "\t Entropy matched: Error distribution matched to the distribution with highest " \
                           "entropy found by FlexConsensus")
        form.addParam('doShowHist', params.LabelParam,
                      label="Display the selected histogram in interactive mode")

    def _getVisualizeDict(self):
        self.chosen = self._choices[self.histChoice.get()]
        return {'doShowHist': self._doShowHist}

    # ------------------- Interactive histogram method -------------------
    def _doShowHist(self, param=None):
        # Load data
        if self.chosen == self._choices[0]:
            data = np.loadtxt(self.protocol._getExtraPath("error_matched_mean.txt"))
        elif self.chosen == self._choices[1]:
            data = np.loadtxt(self.protocol._getExtraPath("error_matched_entropy.txt"))

        # Interactive histogram
        hist = InteractiveHist(data, self.protocol)
        hist.show()

    # ------------------- ------------------- -------------------



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
from pathlib import Path
import os
from scipy.ndimage import gaussian_filter
from xmipp_metadata.image_handler import ImageHandler

from pyworkflow.utils import getExt
from pyworkflow.utils.process import runJob

import flexutils


def getOutputSuffix(protocol, cls):
    """ Get the name to be used for a new output.
    For example: output3DCoordinates7.
    It should take into account previous outputs
    and number with a higher value.
    """
    maxCounter = -1
    for attrName, _ in protocol.iterOutputAttributes(cls):
        suffix = attrName.replace(protocol.OUTPUT_PREFIX, '')
        try:
            counter = int(suffix)
        except:
            counter = 1  # when there is not number assume 1
        maxCounter = max(counter, maxCounter)

    return str(maxCounter + 1) if maxCounter > 0 else ''  # empty if not output


def readZernikeFile(filename):
    with open(filename, 'r') as fid:
        lines = fid.readlines()
    basis_params = np.fromstring(lines[0].strip('\n'), sep=' ')

    z_clnm = []
    for line in lines[1:]:
        z_clnm.append(np.fromstring(line.strip('\n'), sep=' '))
    z_clnm = np.asarray(z_clnm)

    return basis_params, z_clnm


def computeNormRows(array):
    norm = []
    size = int(array.shape[1] / 3)
    for vec in array:
        c_3d = np.vstack([vec[:size], vec[size:2 * size], vec[2 * size:]])
        norm.append(np.linalg.norm(np.linalg.norm(c_3d, axis=1)))
    return np.vstack(norm).flatten()


def getXmippFileName(filename):
    if getExt(filename) == ".mrc":
        filename += ":mrc"
    return filename


def coordsToMap(coords, values, xsize, thr=None):
    indices = coords + np.floor(0.5 * xsize)
    indices = indices.astype(int)

    # Place values on grid
    xsize = int(xsize)
    volume = np.zeros((xsize, xsize, xsize), dtype=np.float32)
    np.add.at(volume, (indices[:, 2], indices[:, 1], indices[:, 0]), values)

    # Filter map
    volume = gaussian_filter(volume, sigma=1.)

    # Volume mask
    if thr is None:
        mask = np.ones(volume.shape)
    else:
        mask = np.zeros(volume.shape)
        mask[volume >= thr] = 1

    return volume, mask


def saveMap(filename, map):
    ImageHandler().write(map, filename, overwrite=True)


def generateVolumesHetSIREN(weigths_file, x_het, outdir, step, architecture):
    args = _getEvalVolArgs(x_het, weigths_file, "het_file", outdir, step=step, architecture=architecture)
    program = flexutils.Plugin.getTensorflowProgram("predict_map_het_siren.py", python=False)
    runJob(None, program, ' '.join(args), numberOfMpi=1)


def generateVolumesDeepNMA(weigths_file, c_nma, outdir, sr, xsize):
    args = _getEvalVolArgs(c_nma, weigths_file, "nma_file", outdir, sr=sr, xsize=xsize)
    program = flexutils.Plugin.getTensorflowProgram("predict_map_deep_nma.py", python=False)
    runJob(None, program, ' '.join(args), numberOfMpi=1)


def _getEvalVolArgs(x_het, weigths_file, x_het_param, outdir, step=None, sr=None, xsize=None, architecture=None):
    if not os.path.exists(outdir):
        os.mkdir(outdir)

    hetFilePath = Path(outdir, "zfile.txt")
    np.savetxt(hetFilePath, x_het)
    hetFilePath = os.path.abspath(os.path.join(outdir, 'zfile.txt'))

    args = ['--weigths_file %s' % weigths_file,
            '--%s %s' % (x_het_param, hetFilePath),
            '--out_path %s' % outdir,
            ]

    if step:
        args.append('--step %d' % step)

    if sr:
        args.append('--sr %f' % sr)

    if xsize:
        args.append('--xsize %d' % xsize)

    if architecture:
        args.append('--architecture %s' % architecture)

    return args



# **************************************************************************
# *
# * Authors:  David Herreros (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os
from pathos.multiprocessing import ProcessingPool as Pool
import subprocess

from pyworkflow.viewer import DESKTOP_TKINTER, WEB_DJANGO, Viewer
from pyworkflow.utils.process import buildRunCommand

from flexutils.protocols.xmipp.protocol_structure_landscape import XmippProtStructureLanscapes
from flexutils.protocols.protocol_dimred import ProtFlexDimRedSpace

import flexutils.constants as const
from flexutils.utils import computeNormRows
import flexutils


class XmippReducedSpaceViewer(Viewer):
    """ Visualize reduced conformational space """
    _label = 'viewer reduced space'
    _targets = [XmippProtStructureLanscapes, ProtFlexDimRedSpace]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]

    def __init__(self, **kwargs):
        Viewer.__init__(self, **kwargs)
        self._data = None

    def getData(self):
        if self._data is None:
            self._data = self.loadData()
        return self._data

    def _visualize(self, obj, **kwargs):
        file_red_space = self.protocol._getExtraPath("red_coords.txt")
        file_z_space = self.protocol._getExtraPath("z_space.txt")
        file_interp = self.protocol._getExtraPath("interp_values.txt")
        particles = self.protocol.outputParticles

        def launchViewerNonBlocking(args):
            (particles, file_red_space, file_z_space, file_interp) = args
            red_space = []
            z_clnm = []
            for particle in particles.iterItems():
                z_clnm.append(particle.getZFlex())
                red_space.append(particle.getZRed())
            z_clnm = np.asarray(z_clnm)
            red_space = np.asarray(red_space)
            if red_space.shape[1] < 3:
                raise Exception("Visualization of spaces with dimension smaller than 3 is not yet implemented. Exiting...")

            # Generate files to call command line
            np.savetxt(file_red_space, red_space)
            # np.savetxt(file_z_space, z_clnm)

            if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
                deformation = computeNormRows(z_clnm)
            else:
                deformation = np.zeros(z_clnm.shape)

            # Generate files to call command line
            np.savetxt(file_interp, deformation)

            # Run slicer
            args = "--data %s --z_space %s --interp_val %s --onlyView" \
                   % (file_red_space, file_z_space, file_interp)
            if hasattr(particles.getFlexInfo(), "umap_weights"):
                args += " --reduce umap --umap_weights %s" % particles.getFlexInfo().getAttr("umap_weights")
            else:
                args += " --reduce pca"
            program = "viewer_interactive_3d.py"
            program = flexutils.Plugin.getProgram(program)

            command = buildRunCommand(program, args, 1)
            subprocess.Popen(command, shell=True)

        # Launch with Pathos
        p = Pool()
        # p.restart()
        p.apipe(launchViewerNonBlocking, args=(particles, file_red_space, file_z_space, file_interp))
        # p.join()
        # p.close()

        return []



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors: Martín Salinas Antón (ssalinasmartin@gmail.com)
# * Authors: David Herreros (dherreros@cnb.csic.es)
# *
# *  BCU, Centro Nacional de Biotecnologia, CSIC
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

# General imports
import os
import numpy as np
from xmipp_metadata.metadata import XmippMetaData
from xmipp_metadata.image_handler import ImageHandler


# Scipion em imports
from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import SetOfParticles, Volume, Integer
from pyworkflow import NEW
from pyworkflow.protocol import params
import pyworkflow.utils as pwutils

# External plugin imports
import xmipp3
from xmipp3.convert import readSetOfImages, rowToParticle


# Protocol output variable name
OUTPUTATTRIBUTE = 'volumeProjections'


def rowToParticleWithLabel(partRow, **kwargs):
    img = rowToParticle(partRow, **kwargs)
    img._xmipp_subtomo_labels = Integer(float(partRow.getValue("subtomo_labels")))
    img.setObjId(None)
    return img


class XmippProtPrepareZernikeVolumes(ProtAnalysis3D, ProtFlexBase):
    """Extracts projections from a set of volumes (keeping a volume id)"""

    # Protocol constants
    _label = 'prepare volumes - Zernike3Deep'
    _devStatus = NEW
    _possibleOutputs = {OUTPUTATTRIBUTE: SetOfParticles}

    # Form constants
    METHOD_FOURIER = 0
    METHOD_REAL_SPACE = 1
    METHOD_SHEARS = 2
    TYPE_N_SAMPLES = 0
    TYPE_STEP = 1

    # --------------------------- Class constructor --------------------------------------------
    def __init__(self, **args):
        # Calling parent class constructor
        super().__init__(**args)

        # Defining execution mode. Steps will take place in parallel now
        # Full tutorial on how to parallelize protocols can be read here:
        # https://scipion-em.github.io/docs/release-3.0.0/docs/developer/parallelization.html
        self.stepsExecutionMode = params.STEPS_PARALLEL

    # --------------------------- DEFINE param functions ------------------------
    def _defineParams(self, form):
        # Defining parallel arguments
        form.addParallelSection(threads=4)

        # Generating form
        form.addSection(label='Input subtomograms')
        form.addParam('inputVolumes', params.PointerParam, pointerClass="SetOfVolumes",
                      label='Set of volumes', help="Set of volumes whose projections will be generated.")
        form.addParam('cleanTmps', params.BooleanParam, default='True', label='Clean temporary files: ', expertLevel=params.LEVEL_ADVANCED,
                        help='Clean temporary files after finishing the execution.\nThis is useful to reduce unnecessary disk usage.')
        form.addParam('transformMethod', params.EnumParam, display=params.EnumParam.DISPLAY_COMBO, default=self.METHOD_FOURIER,
                      choices=['Fourier', 'Real space', 'Shears'], label="Transform method: ", expertLevel=params.LEVEL_ADVANCED,
                      help='Select the algorithm that will be used to obtain the projections.')
        projGroup = form.addGroup('Projection parameters')
        projGroup.addParam('maskMaps', params.BooleanParam, default=True, label="Mask maps?",
                           help="Automatically mask the input volumes to remove noise. Set to false "
                                "if projections do not look sensible enough.")
        projGroup.addParam('typeGeneration', params.EnumParam, display=params.EnumParam.DISPLAY_COMBO, default=self.TYPE_N_SAMPLES,
                           choices=['NSamples', 'Step'], label="Type of sample generation: ",
                           help='Select either the number of samples to be taken or the separation in degrees between each sample.')
        projGroup.addParam('rangeNSamples', params.IntParam, condition='typeGeneration==%d' % self.TYPE_N_SAMPLES, 
                           label='Number of samples:', default=1000,
                           help='Number of samples to be produced.\nIt has to be 1 or greater.')
        projGroup.addParam('rangeStep', params.IntParam, condition='typeGeneration==%d' % self.TYPE_STEP, 
                           label='Step:', default=5,
                           help='Number of degrees each sample will be separated from the next.\nIt has to be greater than 0.')

    # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        # Writing config file
        generateConfigFile = self._insertFunctionStep(self.generateConfigFile)

        # Defining list of function ids to be waited by the createOutput function
        deps = [generateConfigFile]
        # Generating projections for each volume
        volId = 1
        for volume in self.inputVolumes.get():
            deps.append(self._insertFunctionStep(self.generateProjections,
                                                 volume.getFileName(),
                                                 prerequisites=[generateConfigFile]))
            volId += 1

        # Conditionally removing temporary files
        if self.cleanTmps.get():
            deps.append(self._insertFunctionStep(self.removeTempFiles, prerequisites=deps))
        
        # Create output
        self._insertFunctionStep(self.createOutputStep, prerequisites=deps)

    # --------------------------- STEPS functions --------------------------------------------
    def generateConfigFile(self):
        """
        This function writes the config file for Xmipp Phantom.
        """
        confFile = open(self.getXmippParamPath(), "w")

        # Generating file content
        content = '# XMIPP_STAR_1 *\n'
        content += '# Projection Parameters\n'
        content += 'data_block1\n'
        content += '# X and Y projection dimensions [Xdim Ydim]\n'
        content += '_dimensions2D   \'{}\'\n'.format(self.getVolumeDimensions())
        content += '# Rotation range and number of samples [Start Finish NSamples]\n'
        content += '_projRotRange    \'0 360 {}\'\n'.format(self.getStepValue())
        content += '# Rotation angle added noise  [noise (bias)]\n'
        content += '_projRotNoise   \'0\'\n'
        content += '# Tilt range and number of samples for Tilt [Start Finish NSamples]\n'
        content += '_projTiltRange    \'0 180 {}\'\n'.format(self.getStepValue())
        content += '# Tilt angle added noise\n'
        content += '_projTiltNoise   \'0\'\n'
        content += '# Psi range and number of samples\n'
        content += '_projPsiRange    \'0 0 0\'\n'
        content += '# Psi added noise\n_projPsiNoise   \'0\'\n'
        content += '# Noise\n'

        # Writing content to file and closing
        confFile.write(content)
        confFile.close()

    def generateProjections(self, volume):
        """
        This function generates the projection for a given input volume.
        """
        masked_path = self.getVolumeAbsolutePath(volume)
        if self.maskMaps.get():
            vol_path = self.getVolumeAbsolutePath(volume)
            masked_path = self.getVolumeAbsolutePath(self._getExtraPath(pwutils.removeBaseExt(vol_path) + "_masked.mrc"))
            ih = ImageHandler(vol_path)
            mask = ih.generateMask(iterations=50, boxsize=64, smoothStairEdges=False)
            ImageHandler().write(ih.getData() * mask, masked_path, overwrite=True)

        params = '-i {}'.format(masked_path + ":mrc")   # Path to volume
        params += ' -o {}'.format(self.getProjectionAbsolutePath(volume))  # Path to output projection
        params += ' --method {}'.format(self.getMethodValue())             # Projection algorithm
        params += ' --params {}'.format(self.getXmippParamPath())          # Path to Xmipp phantom param file
        self.runJob("xmipp_phantom_project", params, cwd='/home', env=xmipp3.Plugin.getEnviron())

    def removeTempFiles(self):
        """
        This function removes the temporary files of this protocol.
        """
        # Removing Xmipp Phantom config file
        self.runJob('rm', self.getXmippParamPath())

    def createOutputStep(self):
        """
        This function generates the outputs of the protocol.
        """
        # Extracting input
        inputVolumes = self.inputVolumes.get()

        # Creating empty set of particles and setting sampling rate, alignment, and dimensions
        outputSetOfParticles = self._createSetOfParticles()
        outputSetOfParticles.setSamplingRate(inputVolumes.getSamplingRate())
        outputSetOfParticles.setAlignmentProj()
        dimensions = self.getVolumeDimensions().split(' ')
        outputSetOfParticles.setDim((int(dimensions[0]), int(dimensions[1]), 1))

        # Adding projections of each volume as a particle each
        label = 1
        for volume in inputVolumes.iterItems():
            md = XmippMetaData(self.getProjectionMetadataAbsolutePath(volume))
            subtomo_labels = label * np.ones(len(md))
            md[:, "subtomo_labels"] = subtomo_labels
            md.write(self.getProjectionMetadataAbsolutePath(volume), overwrite=True)
            readSetOfImages(self.getProjectionMetadataAbsolutePath(volume), outputSetOfParticles,
                            rowToParticleWithLabel)
            label += 1

        # Defining the ouput with summary and source relation
        outputSetOfParticles.setObjComment(self.getSummary(outputSetOfParticles))
        self._defineOutputs(outputSetOfParticles=outputSetOfParticles)
        self._defineSourceRelation(self.inputVolumes, outputSetOfParticles)

    # --------------------------- INFO functions --------------------------------------------
    def _validate(self):
        """
        This method validates the received params and checks that they all fullfill the requirements needed to run the protocol.
        """
        errors = []

        # Checking if number of samples is greater or equal than 1
        if self.typeGeneration.get() == self.TYPE_N_SAMPLES and (not self.rangeNSamples.get() == None) and self.rangeNSamples.get() < 1:
            errors.append('The number of samples cannot be less than 1.')
        
        # Checking if the step is greater than 0
        if self.typeGeneration.get() == self.TYPE_STEP and (not self.rangeStep.get() == None) and self.rangeStep.get() <= 0:
            errors.append('The step must be greater than 0.')
        
        # Checking if MPI is selected (only threads are allowed)
        if self.numberOfMpi > 1:
            errors.append('MPI cannot be selected, because Scipion is going to drop support for it. Select threads instead.')

        return errors

    def _summary(self):
        """
        This method usually returns a summary of the text provided by '_methods'.
        """
        return []

    def _methods(self):
        """
        This method returns a text intended to be copied and pasted in the paper section 'materials & methods'.
        """
        return []

    # --------------------------- UTILS functions --------------------------------------------
    def scapePath(self, path):
        """
        This function returns the given path with all the spaces in folder names scaped to avoid errors.
        """
        # os.path.baspath adds '\\' when finding a foldername with '\ ', so '\\\' needs to be replaced with ''
        # Then, '\' is inserted before every space again, to include now possible folders with spaces in the absolute path
        return path.replace('\\\ ', ' ').replace(' ', '\ ')
    
    def getVolumeRelativePath(self, volume):
        """
        This method returns a the volume path relative to current directory.
        Path is scaped to support spaces.
        Example:
            if a file is in /home/username/documents/test/import_file.mrc
            and current directory is /home/username/documents
            this will return '/test/import_file.mrc'
        """
        return self.scapePath(volume.getFileName() if isinstance(volume, Volume) else volume)

    def getVolumeAbsolutePath(self, volume):
        """
        This method returns a the absolute path for the volume.
        Path is scaped to support spaces.
        Example: '/home/username/documents/test/import_file.mrc'
        """
        return self.scapePath(os.path.abspath(self.getVolumeRelativePath(volume)))

    def getVolumeName(self, filename):
        """
        This method returns the full name of the given volume files.
        Example: import_file.mrc
        """
        return os.path.basename(filename)
    
    def getCleanVolumeName(self, filename):
        """
        This method returns the full name of the given volume file without the 'import_' prefix.
        Example:
            if filename is 'import_file.mrc'
            this will return 'file.mrc'
        """
        return self.getVolumeName(filename).replace('import_', '')

    def getProjectionName(self, volume):
        """
        This function returns the name of the projection file for a given input volume.
        """
        name = os.path.splitext(self.getCleanVolumeName(self.getVolumeAbsolutePath(volume)))[0]
        return '{}_image.mrc'.format(name)
    
    def getProjectionMetadataName(self, volume):
        """
        This function returns the filename of the metadata file for a given input volume.
        """
        return self.getProjectionName(volume).replace('.mrc', '.xmd')

    def getProjectionAbsolutePath(self, volume):
        """
        This function returns the full path of a given volume.
        """
        return self.scapePath(os.path.abspath(self._getExtraPath(self.getProjectionName(volume))))
    
    def getProjectionMetadataAbsolutePath(self, volume):
        """
        This function returns the full path of a given volume's metadata file.
        """
        return self.getProjectionAbsolutePath(volume).replace('.mrc', '.xmd')

    def getStepValue(self):
        """
        This function translates the provided sample generation input to number of samples for Xmipp phantom.
        """
        if self.typeGeneration.get() == self.TYPE_N_SAMPLES:
            return int(np.sqrt(self.rangeNSamples.get()))
        else:
            # Converting step to number of samples
            return int(np.sqrt(360.0 / self.rangeStep.get()))
    
    def getMethodValue(self):
        """
        This function returns the string value associated to the form value provided by the user regarding transform method.
        """
        if self.transformMethod.get() == self.METHOD_FOURIER:
            return 'fourier'
        elif self.transformMethod.get() == self.METHOD_REAL_SPACE:
            return 'real_space'
        else:
            return 'shears'

    def getVolumeDimensions(self):
        """
        This function retuns the first two dimensions of the subtomograms.
        """
        try:
            dimensions = self.inputVolumes.get().getFirstItem().getDimensions()
            return '{} {}'.format(dimensions[0], dimensions[1])
        except TypeError:
            errorMessage = " ".join(["No subtomograms were received. Check the output of the previous protocol.",
                                     "If you are using the integrated test, run the extract subtomos's test first."])
            raise Exception(errorMessage)

    def getXmippParamPath(self):
        """
        This function returns the path for the config file for Xmipp Phantom.
        """
        return self.scapePath(os.path.abspath(os.path.join(self._getExtraPath(''), 'xmippPhantom.param')))
    
    def getSummary(self, setOfParticles):
        """
        Returns the summary of a given set of particles.
        The summary consists of a text including the number of particles in the set.
        """
        return "Number of projections generated: {}".format(setOfParticles.getSize())



# **************************************************************************
# *
# * Authors:  David Herreros (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os
from pathos.multiprocessing import ProcessingPool as Pool
import subprocess

from pyworkflow.viewer import DESKTOP_TKINTER, WEB_DJANGO, ProtocolViewer
import pyworkflow.protocol.params as params
from pyworkflow.utils.process import buildRunCommand

from flexutils.protocols.xmipp.protocol_angular_alignment_zernike3d import XmippProtAngularAlignmentZernike3D
from flexutils.protocols.xmipp.protocol_focus_zernike3d import XmippProtFocusZernike3D
from flexutils.protocols.xmipp.protocol_reassign_reference_zernike3d import XmippProtReassignReferenceZernike3D
from flexutils.protocols.xmipp.protocol_predict_zernike3deep import TensorflowProtPredictZernike3Deep
from flexutils.protocols.xmipp.protocol_predict_het_siren import TensorflowProtPredictHetSiren
from flexutils.protocols.protocol_score_landscape import ProtFlexScoreLandscape

import flexutils.constants as const
from flexutils.utils import computeNormRows
import flexutils


class XmippLandscapeViewer(ProtocolViewer):
    """ Visualize conformational lanscapes """
    _label = 'viewer conformational landscape'
    _targets = [XmippProtAngularAlignmentZernike3D, XmippProtFocusZernike3D,
                XmippProtReassignReferenceZernike3D, TensorflowProtPredictZernike3Deep,
                TensorflowProtPredictHetSiren, ProtFlexScoreLandscape]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]

    def __init__(self, **kwargs):
        ProtocolViewer.__init__(self, **kwargs)
        self._data = None

    def getData(self):
        if self._data is None:
            self._data = self.loadData()
        return self._data


    def _defineParams(self, form):
        form.addSection(label='Show conformational landscape')
        form.addParam('mode', params.EnumParam, choices=['UMAP', 'PCA'],
                      default=0, display=params.EnumParam.DISPLAY_HLIST,
                      label="Dimensionality reduction method",
                      help="\t * UMAP: usually leads to more meaningful spaces, although execution "
                           "is higher\n"
                           "\t * PCA: faster but less meaningfull spaces \n"
                           "UMAP and PCA are only computed the first time the are used. Afterwards, they "
                           "will be reused to increase performance")
        form.addParam('nb_umap', params.IntParam, label="UMAP neighbors",
                      default=15, condition="mode==0",
                      help="Number of neighbors to associate to each point in the space when computing "
                           "the UMAP space. The higher the number of neighbors, the more predominant "
                           "global in the original space features will be")
        form.addParam('epochs_umap', params.IntParam, label="Number of UMAP epochs",
                      default=1000, condition="mode==0",
                      help="Increasing the number of epochs will lead to more accurate UMAP spaces at the cost "
                           "of larger execution times")
        form.addParam('densmap_umap', params.BooleanParam, label="Compute DENSMAP?",
                      default=False, condition="mode==0",
                      help="DENSMAP will try to bring densities in the UMAP space closer to each other. Execution time "
                           "will increase when computing a DENSMAP")
        form.addParam('threads', params.IntParam, label="Number of threads",
                      default=4, condition="mode==0")
        form.addParam('doShowSpace', params.LabelParam,
                      label="Display the conformational space")

    def _getVisualizeDict(self):
        # self.protocol._createFilenameTemplates()
        return {'doShowSpace': self._doShowSpace}

    def _doShowSpace(self, param=None):
        particles = self.protocol.outputParticles
        file_z_space = self.protocol._getExtraPath("z_space.txt")
        file_interp = self.protocol._getExtraPath("interp_values.txt")
        mode = self.mode.get()

        if mode == 0:
            file_coords = self.protocol._getExtraPath("umap_coords.txt")
            mode = [mode, self.nb_umap.get(), self.epochs_umap.get(), self.threads.get(), self.densmap_umap.get()]
        elif mode == 1:
            file_coords = self.protocol._getExtraPath("pca_coords.txt")
            mode = [mode, ]

        def launchViewerNonBlocking(args):
            (particles, file_z_space, file_interp, file_coords, mode) = args

            z_clnm = []
            for particle in particles.iterItems():
                z_clnm.append(particle.getZFlex())
            z_clnm = np.asarray(z_clnm)
            if z_clnm.shape[1] < 3:
                raise Exception("Visualization of spaces with dimension smaller than 3 is not yet implemented. Exiting...")

            # Generate files to call command line
            np.savetxt(file_z_space, z_clnm)

            # Compute/Read UMAP or PCA
            if mode[0] == 0:
                if not os.path.isfile(file_coords):
                    args = "--input %s --umap --output %s --n_neighbors %d " \
                           "--n_epochs %d --n_components 3 --thr %d" \
                           % (file_z_space, file_coords, mode[1], mode[2], mode[3])
                    if mode[4]:
                        args += " --densmap"
                    program = "dimensionality_reduction.py"
                    program = flexutils.Plugin.getProgram(program)
                    command = buildRunCommand(program, args, 1)
                    p = subprocess.Popen(command, shell=True)
                    p.wait()
            elif mode[0] == 1:
                if not os.path.isfile(file_coords):
                    args = "--input %s --pca --output %s --n_components 3" % (file_z_space, file_coords)
                    program = "dimensionality_reduction.py"
                    program = flexutils.Plugin.getProgram(program)
                    command = buildRunCommand(program, args, 1)
                    p = subprocess.Popen(command, shell=True)
                    p.wait()

            if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
                deformation = computeNormRows(z_clnm)
            else:
                deformation = np.zeros(z_clnm.shape)

            # Generate files to call command line
            np.savetxt(file_interp, deformation)

            # Run slicer
            args = "--data %s --z_space %s --interp_val %s --onlyView" \
                   % (file_coords, file_z_space, file_interp)
            program = "viewer_interactive_3d.py"
            program = flexutils.Plugin.getProgram(program)

            command = buildRunCommand(program, args, 1)
            subprocess.Popen(command, shell=True)

        # Launch with Pathos
        p = Pool()
        # p.restart()
        p.apipe(launchViewerNonBlocking, args=(particles, file_z_space, file_interp, file_coords, mode))
        # p.join()
        # p.close()

        return []



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

import xmipp3
from pyworkflow import NEW
from pyworkflow.protocol.params import PointerParam

from pwem.protocols import ProtAnalysis3D
from pwem.objects import SetOfParticlesFlex

import flexutils
import flexutils.constants as const

import pwem.emlib.metadata as md
from xmipp3.convert import writeSetOfImages, imageToRow, coordinateToRow

from flexutils.utils import getXmippFileName


class XmippProtStatisticsZernike3D(ProtAnalysis3D):
    """ 3D visualization of motion statistics """

    _label = 'motion statistics - Zernike3D'
    _devStatus = NEW
    OUTPUT_PREFIX = 'zernike3DClasses'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('particles', PointerParam, label="Zernike3D particles",
                      pointerClass='SetOfParticlesFlex', important=True,
                      help="Particles must have a set of Zernike3D coefficients associated")
        form.addParam('structure', PointerParam, label="Atomic structure",
                      pointerClass="AtomStruct", allowsNull=True,
                      help="Optional: If provided, it will be possible to visualize "
                           "the PDB when analyzing the motion statistics")
        form.addParallelSection(threads=4, mpi=0)

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.launchVolumeViewer, interactive=True)

    # --------------------------- STEPS functions -----------------------------
    def launchVolumeViewer(self):
        particles = self.particles.get()
        mask = particles.getFlexInfo().refMask.get()

        L1 = particles.getFlexInfo().L1.get()
        L2 = particles.getFlexInfo().L2.get()
        z_clnm_vec = {}
        # def_vec = {}

        for particle in particles.iterItems():
            z_clnm = particle.getZFlex()
            z_clnm_vec[particle.getObjId()] = z_clnm.reshape(-1)
            # def_vec[particle.getObjId()] = particle._xmipp_sphDeformation.get()

        def zernikeRow(part, partRow, **kwargs):
            imageToRow(part, partRow, md.MDL_IMAGE, **kwargs)
            coord = part.getCoordinate()
            idx = part.getObjId()
            if coord is not None:
                coordinateToRow(coord, partRow, copyId=False)
            if part.hasMicId():
                partRow.setValue(md.MDL_MICROGRAPH_ID, int(part.getMicId()))
                partRow.setValue(md.MDL_MICROGRAPH, str(part.getMicId()))
            partRow.setValue(md.MDL_SPH_COEFFICIENTS, z_clnm_vec[idx].tolist())
            # partRow.setValue(md.MDL_SPH_DEFORMATION, def_vec[idx])

        if not os.path.isfile(self._getExtraPath("particles.xmd")):
            writeSetOfImages(particles, self._getExtraPath("particles.xmd"), zernikeRow)

        # Run viewer
        args = "--i %s --r %s --L1 %d --L2 %d --sr %f --thr %d" \
               % (self._getExtraPath("particles.xmd"), getXmippFileName(mask), L1, L2,
                  particles.getSamplingRate(), self.numberOfThreads.get())
        if self.structure.get():
            args += " --atom_file %s" % self.structure.get().getFileName()
        program = "viewer_map.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args)

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        pass

    def _methods(self):
        return [
            "3D visualization of motion statistics",
        ]

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        particles = self.particles.get()
        if isinstance(particles, SetOfParticlesFlex):
            if particles.getFlexInfo().getProgName() != const.ZERNIKE3D:
                errors.append("The flexibility information associated with the particles is not "
                              "coming from the Zernike3D algorithm. Please, provide a set of particles "
                              "with the correct flexibility information.")
        return errors



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreos@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import numpy as np
import re
from xmipp_metadata.metadata import XmippMetaData
from xmipp_metadata.image_handler import ImageHandler

import pyworkflow.protocol.params as params
from pyworkflow.object import Integer, Float, String, CsvList, Boolean
from pyworkflow.utils.path import moveFile
from pyworkflow import VERSION_2_0

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
import pwem.emlib.metadata as md
from pwem.constants import ALIGN_PROJ
from pwem.objects import ParticleFlex, SetOfParticlesFlex

from xmipp3.convert import createItemMatrix, setXmippAttributes, writeSetOfParticles, \
    geometryFromMatrix, matrixFromGeometry
import xmipp3

import flexutils
import flexutils.constants as const
from flexutils.utils import getXmippFileName
from flexutils.protocols.xmipp.utils.custom_pdb_parser import PDBUtils


class TensorflowProtPredictZernike3Deep(ProtAnalysis3D, ProtFlexBase):
    """ Predict Zernike3D coefficents for a set of particles based on a trained
     Zernike3Deep network. """
    _label = 'predict - Zernike3Deep'
    _lastUpdateVersion = VERSION_2_0

    # --------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addHidden(params.USE_GPU, params.BooleanParam, default=True,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                             Select the one you want to use.")
        form.addHidden(params.GPU_LIST, params.StringParam, default='0',
                       expertLevel=params.LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")
        group = form.addGroup("Data")
        group.addParam('inputParticles', params.PointerParam, label="Input particles to predict",
                       pointerClass='SetOfParticles')
        group.addParam('zernikeProtocol', params.PointerParam, label="Zernike3Deep trained network",
                       pointerClass='TensorflowProtAngularAlignmentZernike3Deep',
                       help="Previously executed 'angular align - Zernike3Deep'. "
                            "This will allow to load the network trained in that protocol to be used during "
                            "the prediction")
        group = form.addGroup("Mask type")
        group.addParam('convertBinary', params.BooleanParam, default=False,
                       label="Associate field to binary mask?",
                       help="If a regions mask has been used to trained the Zernike3Deep network, you "
                            "can set this parameter to yes to reassociate the deformation coefficients "
                            "to a binary mask generated from the regions mask.")
        form.addParallelSection(threads=4, mpi=0)

    def _createFilenameTemplates(self):
        """ Centralize how files are called """
        myDict = {
            'imgsFn': self._getExtraPath('input_particles.xmd'),
            'fnVol': self._getExtraPath('volume.mrc'),
            'fnVolMask': self._getExtraPath('mask.mrc'),
            'fnStruct': self._getExtraPath('structure.txt'),
            'fnConnect': self._getExtraPath("connectivity.txt"),
            'fnCA': self._getExtraPath("ca_indices.txt"),
            'fnOutDir': self._getExtraPath()
        }
        self._updateFilenamesDict(myDict)

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._createFilenameTemplates()
        self._insertFunctionStep(self.writeMetaDataStep)
        self._insertFunctionStep(self.predictStep)
        if self.convertBinary:
            self._insertFunctionStep(self.convertBinaryStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions -----------------------
    def writeMetaDataStep(self):
        imgsFn = self._getFileName('imgsFn')
        fnVol = self._getFileName('fnVol')
        fnVolMask = self._getFileName('fnVolMask')
        structure = self._getFileName('fnStruct')
        md_file = self._getFileName('imgsFn')

        inputParticles = self.inputParticles.get()
        zernikeProtocol = self.zernikeProtocol.get()
        Xdim = inputParticles.getXDim()
        self.newXdim = zernikeProtocol.boxSize.get()
        i_sr = 1. / inputParticles.getSamplingRate()

        ih = ImageHandler()
        inputVolume = zernikeProtocol.inputVolume.get().getFileName()
        ih.convert(getXmippFileName(inputVolume), fnVol)
        curr_vol_dim = ImageHandler(getXmippFileName(inputVolume)).getDimensions()[-1]
        if curr_vol_dim != self.newXdim:
            self.runJob("xmipp_image_resize",
                        "-i %s --dim %d " % (fnVol, self.newXdim), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

        inputMask = zernikeProtocol.inputVolumeMask.get().getFileName()
        if inputMask:
            ih.convert(getXmippFileName(inputMask), fnVolMask)
            curr_mask_dim = ImageHandler(getXmippFileName(inputMask)).getDimensions()[-1]
            if curr_mask_dim != self.newXdim:
                self.runJob("xmipp_image_resize",
                            "-i %s --dim %d --interp nearest" % (fnVolMask, self.newXdim), numberOfMpi=1,
                            env=xmipp3.Plugin.getEnviron())

        if zernikeProtocol.referenceType.get() == 1:  # Structure reference
            inputVolume = zernikeProtocol.inputVolume.get().getFileName()
            inputStruct = zernikeProtocol.inputStruct.get().getFileName()
            structure_file = self._getFileName('fnStruct')
            connect_file = self._getFileName('fnConnect')
            ca_file = self._getFileName('fnCA')
            parser = PDBUtils(selectionString=zernikeProtocol._subset[zernikeProtocol.atomSubset.get()])
            pdb_coordinates, ca_indices, connectivity = parser.parsePDB(inputStruct)
            pdb_coordinates *= i_sr
            ih = ImageHandler(getXmippFileName(inputVolume))
            vol = ih.getData()
            factor = 0.5 * ih.getDimensions()[-1]
            pdb_indices = np.round(pdb_coordinates + factor).astype(int)
            values = vol[pdb_indices[:, 2], pdb_indices[:, 1], pdb_indices[:, 0]]
            pdb_coordinates = np.c_[pdb_coordinates, values]
            np.savetxt(structure_file, pdb_coordinates)
            np.savetxt(connect_file, connectivity)
            np.savetxt(ca_file, ca_indices)

        writeSetOfParticles(inputParticles, imgsFn)

        # Write extra attributes (if needed)
        md = XmippMetaData(md_file)
        if isinstance(inputParticles, SetOfParticlesFlex) and \
                inputParticles.getFlexInfo().getProgName() == const.ZERNIKE3D:
            z_space = np.asarray([particle.getZFlex() for particle in inputParticles.iterItems()])
            md[:, "zernikeCoefficients"] = (Xdim / self.newXdim) * z_space
        if hasattr(inputParticles.getFirstItem(), "_xmipp_subtomo_labels"):
            labels = np.asarray([int(particle._xmipp_subtomo_labels) for particle in inputParticles.iterItems()])
            md[:, "subtomo_labels"] = labels
        md.write(md_file, overwrite=True)

        if self.newXdim != Xdim:
            params = "-i %s -o %s --save_metadata_stack %s --fourier %d" % \
                     (imgsFn,
                      self._getTmpPath('scaled_particles.stk'),
                      self._getExtraPath('scaled_particles.xmd'),
                      self.newXdim)
            if self.numberOfMpi.get() > 1:
                params += " --mpi_job_size %d" % int(inputParticles.getSize() / self.numberOfMpi.get())
            self.runJob("xmipp_image_resize", params, numberOfMpi=self.numberOfMpi.get(),
                        env=xmipp3.Plugin.getEnviron())
            moveFile(self._getExtraPath('scaled_particles.xmd'), imgsFn)

    def predictStep(self):
        zernikeProtocol = self.zernikeProtocol.get()
        md_file = self._getFileName('imgsFn')
        weigths_file = zernikeProtocol._getExtraPath(os.path.join('network', 'zernike3deep_model.h5'))
        L1 = zernikeProtocol.l1.get()
        L2 = zernikeProtocol.l2.get()
        pad = zernikeProtocol.pad.get()
        correctionFactor = self.inputParticles.get().getXDim() / zernikeProtocol.boxSize.get()
        sr = correctionFactor * self.inputParticles.get().getSamplingRate()
        applyCTF = zernikeProtocol.applyCTF.get()
        args = "--md_file %s --weigths_file %s --L1 %d --L2 %d " \
               "--pad %d --sr %f --apply_ctf %d" \
               % (md_file, weigths_file, L1, L2, pad, sr, applyCTF)

        if zernikeProtocol.refinePose.get():
            args += " --refine_pose"

        if zernikeProtocol.ctfType.get() == 0:
            args += " --ctf_type apply"
        elif zernikeProtocol.ctfType.get() == 1:
            args += " --ctf_type wiener"

        if zernikeProtocol.architecture.get() == 0:
            args += " --architecture convnn"
        elif zernikeProtocol.architecture.get() == 1:
            args += " --architecture mlpnn"

        if self.useGpu.get():
            gpu_list = ','.join([str(elem) for elem in self.getGpuList()])
            args += " --gpu %s" % gpu_list
        program = flexutils.Plugin.getTensorflowProgram("predict_zernike3deep.py", python=False)
        self.runJob(program, args, numberOfMpi=1)

    def convertBinaryStep(self):
        zernikeProtocol = self.zernikeProtocol.get()
        maskRegOri = zernikeProtocol.inputVolumeMask.get().getFileName()
        maskReg = self._getFileName('fnVolMask')
        maskBin = self._getExtraPath("binary_mask.mrc")
        maskBinOri = self._getExtraPath("binary_mask_ori.mrc")
        L1 = zernikeProtocol.l1.get()
        L2 = zernikeProtocol.l2.get()
        md_file = self._getFileName('imgsFn')

        # Convert mask to binary (original)
        data = ImageHandler(maskRegOri).getData()
        data[data > 0] = 1.0
        ImageHandler().write(data, maskBinOri, overwrite=True)

        # Convert mask to binary (downsampled)
        data = ImageHandler(maskReg).getData()
        data[data > 0] = 1.0
        boxsize = data.shape[0]
        ImageHandler().write(data, maskBin, overwrite=True)

        # Update deformation field
        args = "--md_file %s --mask_reg %s --mask_bin %s --boxsize %d --l1 %d --l2 %d --thr %d" \
               % (md_file, maskReg, maskBin, boxsize, L1, L2, self.numberOfThreads.get())
        program = "field_regions_to_binary_zernike3d.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args, env=xmipp3.Plugin.getEnviron())

    def createOutputStep(self):
        inputParticles = self.inputParticles.get()
        zernikeProtocol = self.zernikeProtocol.get()
        Xdim = inputParticles.getXDim()
        self.newXdim = zernikeProtocol.boxSize.get()
        model_path = zernikeProtocol._getExtraPath(os.path.join('network', 'zernike3deep_model.h5'))
        md_file = self._getFileName('imgsFn')

        metadata = XmippMetaData(md_file)
        zernike_space = np.asarray([np.fromstring(item, sep=',') for item in metadata[:, 'zernikeCoefficients']])

        if metadata.isMetaDataLabel('delta_angle_rot'):
            delta_rot = metadata[:, 'delta_angle_rot']
            delta_tilt = metadata[:, 'delta_angle_tilt']
            delta_psi = metadata[:, 'delta_angle_psi']
            delta_shift_x = metadata[:, 'delta_shift_x']
            delta_shift_y = metadata[:, 'delta_shift_y']

        refinePose = zernikeProtocol.refinePose.get()

        inputSet = self.inputParticles.get()
        partSet = self._createSetOfParticlesFlex(progName=const.ZERNIKE3D)
        partSet.setHasCTF(inputSet.hasCTF())

        partSet.copyInfo(inputSet)
        partSet.getFlexInfo().setProgName(const.ZERNIKE3D)
        partSet.setAlignmentProj()

        correctionFactor = Xdim / self.newXdim
        zernike_space = correctionFactor * zernike_space

        inverseTransform = partSet.getAlignment() == ALIGN_PROJ

        idx = 0
        for particle in inputSet.iterItems():
            # z = correctionFactor * zernike_space[idx]

            outParticle = ParticleFlex(progName=const.ZERNIKE3D)
            outParticle.copyInfo(particle)
            outParticle.getFlexInfo().setProgName(const.ZERNIKE3D)

            outParticle.setZFlex(zernike_space[idx])

            if refinePose:
                tr_ori = particle.getTransform().getMatrix()
                shifts, angles = geometryFromMatrix(tr_ori, inverseTransform)

                # Apply delta angles
                angles[0] += delta_rot[idx]
                angles[1] += delta_tilt[idx]
                angles[2] += delta_psi[idx]

                # Apply delta shifts
                shifts[0] += correctionFactor * delta_shift_x[idx]
                shifts[1] += correctionFactor * delta_shift_y[idx]

                # Set new transformation matrix
                tr = matrixFromGeometry(shifts, angles, inverseTransform)
                outParticle.getTransform().setMatrix(tr)

            partSet.append(outParticle)

            idx += 1

        partSet.getFlexInfo().L1 = Integer(zernikeProtocol.l1.get())
        partSet.getFlexInfo().L2 = Integer(zernikeProtocol.l2.get())
        partSet.getFlexInfo().Rmax = Float(Xdim / 2)
        partSet.getFlexInfo().modelPath = String(model_path)

        inputMask = self._getExtraPath("binary_mask_ori.mrc") if self.convertBinary.get() \
            else zernikeProtocol.inputVolumeMask.get().getFileName()
        inputVolume = zernikeProtocol.inputVolume.get().getFileName()
        partSet.getFlexInfo().refMask = String(inputMask)
        partSet.getFlexInfo().refMap = String(inputVolume)

        partSet.getFlexInfo().refPose = refinePose

        self._defineOutputs(outputParticles=partSet)
        self._defineTransformRelation(self.inputParticles, partSet)

    # --------------------------- UTILS functions -----------------------
    def _updateParticle(self, item, row):
        setXmippAttributes(item, row, md.MDL_ANGLE_ROT, md.MDL_ANGLE_TILT,
                           md.MDL_ANGLE_PSI, md.MDL_SHIFT_X, md.MDL_SHIFT_Y,
                           md.MDL_FLIP, md.MDL_SPH_DEFORMATION,
                           md.MDL_SPH_COEFFICIENTS)
        createItemMatrix(item, row, align=ALIGN_PROJ)

    def getInputParticles(self):
        return self.inputParticles.get()

    def readPDB(self, fnIn):
        with open(fnIn) as f:
            lines = f.readlines()
        return lines

    def PDB2List(self, lines):
        zernikeProtocol = self.zernikeProtocol.get()
        newlines = []
        for line in lines:
            eval = re.search(r'^ATOM\s+\d+\s+/N|CA|C|O/\s+', line) \
                if zernikeProtocol.onlyBackbone.get() else line.startswith("ATOM ")
            if eval:
                try:
                    x = float(line[30:38])
                    y = float(line[38:46])
                    z = float(line[46:54])
                    val = float(line[54:60])
                    newline = [x, y, z, val]
                    newlines.append(newline)
                except:
                    pass
        return newlines

    # ----------------------- VALIDATE functions -----------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        return errors
